
package main

import (
	"fmt"
	"sync"
	"time"
)

// we still have a shared buffer and we are still using mutex lock 
// to prevent race conditions on that share data
//
// but here we use a different approach to make the main function
// wait for the two ping pong goroutines to finish
// recall that since main spawns the goroutines, then when main 
// ends any goroutines that are still running will end as well 
//
// we use a WaitGroup, which is a data structure in the sync package 
// a WaitGroup is just a nicely packaged counter that the goroutines
// are given access to via parameter passing by reference


var sharedData string // Shared variable buffer for communication
var mu sync.Mutex     // Mutex to synchronize access to sharedData
const pps = 10

func ping(wg *sync.WaitGroup) {  // type of wg is declared to be a pointer to
                                 // a WaitGroup
                                 // compiler will automatically dereference wg
                                 // when it is used, like wg.Done()
   defer wg.Done()  // very cool feature in go
                    // defer means this operation is postponed 
                    // until the rest of the call to ping is done
                    // and it is about to return (either normally or error exit)
                    // at that point the wg.Done() func call is executed
                    // wg.Done decrements the counter in this WaitGroup
   for i := 0; i < pps; i++ { // Simulate 5 ping-pong exchanges
      mu.Lock()
      if sharedData == "pong" { // Wait for pong to be written
         fmt.Println("(",i,") Ping received:", sharedData)
         sharedData = "ping"
         fmt.Println("Ping sent:", sharedData)
      }
      mu.Unlock()
      time.Sleep(500 * time.Millisecond) // Simulate processing time
   }
}

func pong(wg *sync.WaitGroup) {
   defer wg.Done()  // note that putting wg.Done() in this function implies
                    // that the function is being written to be the code of
                    // a goroutine
   for i := 0; i < pps; i++ { // Simulate 5 ping-pong exchanges
      mu.Lock()
      if sharedData == "ping" || sharedData == "" { // Wait for ping to be written 
                                                    // or initial state
         fmt.Println("(",i,") Pong received:", sharedData)
         sharedData = "pong"
         fmt.Println("Pong sent: ", sharedData)
      }
      mu.Unlock()
      time.Sleep(500 * time.Millisecond) // Simulate processing time
   }
}

func main() {
   // WaitGroup to ensure both goroutines finish
   // the one WaitGroup is not a global variable
   // rather it will be passed in to each goroutine as a parameter
   // in this way it is reminiscent of the design pattern observer
   var wg sync.WaitGroup

   // Add 2 goroutines to the wait group
   wg.Add(2)

   // Start the ping and pong goroutines
   go ping(&wg) // wg is passed by reference
   go pong(&wg)

   // Wait for both goroutines to finish
   wg.Wait()  // main is blocked until the count in wb falls to 0
}

